import logging
import os
import shutil
import tarfile
import tempfile
from pathlib import Path
from typing import IO, Any, Tuple, Union
from urllib.parse import urlparse
from zipfile import ZipFile, is_zipfile

import requests
from requests.adapters import HTTPAdapter
from tqdm import tqdm
from urllib3.util.retry import Retry

logger = logging.getLogger(__name__)


def remove_file_or_directory(path: Union[str, Path]) -> None:
    try:
        path = Path(path)
        if path.is_dir():
            shutil.rmtree(path)
        else:
            os.remove(path)
    except FileNotFoundError:
        pass


def is_archive_file(filename: Union[str, Path]) -> bool:
    if not Path(filename).is_file():
        return False
    return is_zipfile(filename) or tarfile.is_tarfile(filename)


def extract_archive_file(
    source_path: Union[str, Path], target_path: Union[str, Path]
) -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        if is_zipfile(source_path):
            with ZipFile(source_path, "r") as zip_file:
                zip_file.extractall(temp_dir)
        elif tarfile.is_tarfile(source_path):
            with tarfile.open(source_path) as tar_file:
                tar_file.extractall(temp_dir)
        os.replace(temp_dir, target_path)


def extract_path(filename: Union[str, Path]) -> str:
    parsed = urlparse(str(filename))
    return parsed.path


def is_local(url_or_filename: Union[str, Path]) -> bool:
    if isinstance(url_or_filename, Path):
        return True

    parsed = urlparse(str(url_or_filename))
    if parsed.scheme in ("", "file", "osfs"):
        return True

    return False


def get_parent_path_and_filename(path: Union[str, Path]) -> Tuple[str, str]:
    if isinstance(path, Path):
        parent = str(path.parent)
        name = str(path.name)
        return parent, name

    parsed_url = urlparse(path)
    scheme = parsed_url.scheme
    netloc = parsed_url.netloc
    path = parsed_url.path

    splitted = str(path).rsplit("/", 1)
    if len(splitted) == 2:
        parent, name = splitted
    else:
        parent = ""
        name = str(path)

    if scheme and netloc:
        parent = f"{scheme}://{netloc}/{parent}"

    return parent, name


def sizeof_fmt(num: Union[int, float], suffix: str = "B") -> str:
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1024.0:
            return "%3.1f %s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f %s%s" % (num, "Yi", suffix)


def http_get(url: str, temp_file: IO[Any]) -> None:
    with _session_with_backoff() as session:
        req = session.get(url, stream=True)
        req.raise_for_status()
        content_length = req.headers.get("Content-Length")
        total = int(content_length) if content_length is not None else None
        progress = tqdm(unit="B", total=total, desc="downloading")
        for chunk in req.iter_content(chunk_size=1024):
            if chunk:  # filter out keep-alive new chunks
                progress.update(len(chunk))
                temp_file.write(chunk)
        progress.close()


def _session_with_backoff() -> requests.Session:
    session = requests.Session()
    retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
    session.mount("http://", HTTPAdapter(max_retries=retries))
    session.mount("https://", HTTPAdapter(max_retries=retries))

    return session



from contextlib import contextmanager
from pathlib import Path
from typing import IO, Any, Iterator, Optional, Union

from minato.cache import Cache
from minato.config import Config
from minato.filesystems import FileSystem
from minato.minato import Minato

__version__ = "0.3.0"
__all__ = [
    "Cache",
    "Config",
    "FileSystem",
    "Minato",
    "cached_path",
    "download",
    "open",
    "upload",
]


@contextmanager
def open(
    url_or_filename: Union[str, Path],
    mode: str = "r",
    extract: bool = False,
    use_cache: bool = True,
    force_download: bool = False,
    force_extract: bool = False,
    cache_root: Optional[Union[str, Path]] = None,
    expire_days: Optional[int] = None,
) -> Iterator[IO[Any]]:
    config = Config.load(
        cache_root=cache_root,
        expire_days=expire_days,
    )

    with Minato(config).open(
        url_or_filename,
        mode=mode,
        extract=extract,
        use_cache=use_cache,
        force_download=force_download,
        force_extract=force_extract,
    ) as fp:
        yield fp


def cached_path(
    url_or_filename: Union[str, Path],
    extract: bool = False,
    force_download: bool = False,
    force_extract: bool = False,
    retry: bool = True,
    cache_root: Optional[Union[str, Path]] = None,
    expire_days: Optional[int] = None,
) -> Path:
    config = Config.load(
        cache_root=cache_root,
        expire_days=expire_days,
    )

    return Minato(config).cached_path(
        url_or_filename,
        extract=extract,
        force_download=force_download,
        force_extract=force_extract,
        retry=retry,
    )


def download(url: str, filename: Union[str, Path]) -> None:
    filename = Path(filename)
    Minato.download(url, filename)


def upload(filename: Union[str, Path], url: str) -> None:
    filename = Path(filename)
    Minato.upload(filename, url)



import os
import re
import tempfile
import threading
from contextlib import contextmanager
from pathlib import Path
from typing import IO, Any, Iterator, Union

import boto3

from minato.filesystems.filesystem import FileSystem


@FileSystem.register(["s3"])
class S3FileSystem(FileSystem):
    def __init__(self, url_or_filename: Union[str, Path]) -> None:
        super().__init__(url_or_filename)
        self._tlocal = threading.local()

        self._bucket_name = self._url.hostname or ""
        self._key = re.sub(r"^/", "", self._url.path)

        self._aws_access_key_id = self._url.username or os.environ.get(
            "AWS_ACCESS_KEY_ID"
        )
        self._aws_secret_access_key = self._url.password or os.environ.get(
            "AWS_SECRET_ACCESS_KEY"
        )
        self._endpoint_url = self._url.get_query("endpoint_url")
        self._region_name = self._url.get_query("region")

    def _get_resource(self):  # type: ignore
        if not hasattr(self._tlocal, "resource"):
            self._tlocal.resource = boto3.resource(
                "s3",
                region_name=self._region_name,
                aws_access_key_id=self._aws_access_key_id,
                aws_secret_access_key=self._aws_secret_access_key,
                endpoint_url=self._endpoint_url,
            )
        return self._tlocal.resource

    def _get_client(self):  # type: ignore
        if not hasattr(self._tlocal, "client"):
            self._tlocal.client = boto3.client(
                "s3",
                region_name=self._region_name,
                aws_access_key_id=self._aws_access_key_id,
                aws_secret_access_key=self._aws_secret_access_key,
                endpoint_url=self._endpoint_url,
            )
        return self._tlocal.client

    def _download_fileobj(self, key: str, fp: IO[Any]) -> None:
        client = self._get_client()  # type: ignore
        client.download_fileobj(self._bucket_name, key, fp)

    def _upload_fileobj(self, fp: IO[Any], key: str) -> None:
        client = self._get_client()  # type: ignore
        client.upload_fileobj(fp, self._bucket_name, key)

    def exists(self) -> bool:
        resource = self._get_resource()  # type: ignore
        bucket = resource.Bucket(self._bucket_name)
        objects = list(bucket.objects.filter(Prefix=self._key))
        return len(objects) > 0

    def download(self, path: Union[str, Path]) -> None:
        if not self.exists():
            raise FileNotFoundError(self._url.raw)

        if isinstance(path, str):
            path = Path(path)

        resource = self._get_resource()  # type: ignore
        bucket = resource.Bucket(self._bucket_name)
        objects = list(bucket.objects.filter(Prefix=self._key))

        if len(objects) == 1:  # if the given path is a file
            obj = objects[0]
            if path.is_dir():
                file_path = path / os.path.basename(obj.key)
            else:
                file_path = path
            os.makedirs(file_path.parent, exist_ok=True)
            bucket.download_file(obj.key, str(file_path))
        else:  # if the given path is a directory
            for obj in objects:
                relpath = os.path.relpath(obj.key, self._key)
                parent_dir = path / os.path.dirname(relpath)
                os.makedirs(parent_dir, exist_ok=True)

                file_path = path / relpath
                bucket.download_file(obj.key, str(file_path))

    @contextmanager
    def open_file(
        self,
        mode: str = "r",
    ) -> Iterator[IO[Any]]:
        if "x" in mode and self.exists():
            raise FileExistsError(self._url.raw)

        local_file = tempfile.NamedTemporaryFile(delete=False)
        try:
            if "r" in mode or "a" in mode or "+" in mode:
                if not self.exists():
                    raise FileNotFoundError(self._url.raw)
                self._download_fileobj(self._key, local_file)

            local_file.close()
            with open(local_file.name, mode) as fp:
                yield fp

            if "w" in mode or "a" in mode or "+" in mode or "x" in mode:
                with open(local_file.name, "rb") as fp:
                    self._upload_fileobj(fp, self._key)

        finally:
            local_file.close()
            os.remove(local_file.name)



import shutil
from contextlib import contextmanager
from pathlib import Path
from typing import IO, Any, Iterator, Union

from minato.filesystems.filesystem import FileSystem


@FileSystem.register(["file", "osfs", ""])
class OSFileSystem(FileSystem):
    def __init__(self, url_or_filename: Union[str, Path]) -> None:
        super().__init__(url_or_filename)
        self._path = Path(self._url.path)

    def exists(self) -> bool:
        return self._path.exists()

    def download(self, path: Union[str, Path]) -> None:
        shutil.copy(self._path, path)

    @contextmanager
    def open_file(
        self,
        mode: str = "r",
    ) -> Iterator[IO[Any]]:
        with self._path.open(mode) as fp:
            yield fp



import argparse
from pathlib import Path

from minato.commands.subcommand import Subcommand
from minato.config import Config
from minato.minato import Minato
from minato.table import Table
from minato.util import is_archive_file, sizeof_fmt


@Subcommand.register(
    "list",
    description="show list of cached files",
    help="show list of cached files",
)
class ListCommand(Subcommand):
    def set_arguments(self) -> None:
        self.parser.add_argument("uid_or_url", nargs="*", default=[])
        self.parser.add_argument("--sort", type=str, default=None)
        self.parser.add_argument("--desc", action="store_true")
        self.parser.add_argument("--details", action="store_true")
        self.parser.add_argument("--column-width", type=int, default=None)
        self.parser.add_argument("--root", type=Path, default=None)
        self.parser.add_argument("--expired", action="store_true", default=None)
        self.parser.add_argument("--failed", action="store_true", default=None)
        self.parser.add_argument("--completed", action="store_true", default=None)
        self.parser.add_argument("--expire-days", type=int, default=None)

    def run(self, args: argparse.Namespace) -> None:
        config = Config.load(
            cache_root=args.root,
            expire_days=args.expire_days,
        )
        minato = Minato(config)
        cache = minato.cache

        cached_files = cache.filter(
            queries=args.uid_or_url,
            expired=args.expired or args.expire_days is not None,
            failed=args.failed,
            completed=args.completed,
        )

        columns = ["uid", "url", "size", "type", "status", "expired"]
        if args.details:
            columns.append("local_path")
            columns.append("created_at")
            columns.append("updated_at")
            columns.append("extraction_path")

        table = Table(
            columns=columns,
            max_column_width=args.column_width,
        )

        for cached_file in cached_files:
            info = cached_file.to_dict()

            info["uid"] = info["uid"][:8]

            if cached_file.local_path.exists():
                info["size"] = sizeof_fmt(cached_file.local_path.stat().st_size)
            else:
                info["size"] = "-"

            info["type"] = get_cache_type(cached_file.local_path)
            info["expired"] = cache.is_expired(cached_file)

            table.add(info)

        if args.sort:
            table.sort(args.sort, args.desc)

        table.print()


def get_cache_type(path: Path) -> str:
    if path.is_dir():
        return "dir"
    if is_archive_file(path):
        return "archive"
    return "file"



from minato.filesystems.filesystem import FileSystem, download, open_file  # noqa: F401
from minato.filesystems.gcs import GCSFileSystem  # noqa: F401
from minato.filesystems.http import HttpFileSystem  # noqa: F401
from minato.filesystems.osfs import OSFileSystem  # noqa: F401
from minato.filesystems.s3 import S3FileSystem  # noqa: F401



from contextlib import contextmanager
from pathlib import Path
from typing import IO, Any, Iterator, Optional, Union

from minato.cache import Cache, CacheStatus
from minato.config import Config
from minato.exceptions import InvalidCacheStatus
from minato.filesystems import download, open_file
from minato.util import (
    extract_archive_file,
    extract_path,
    is_archive_file,
    is_local,
    remove_file_or_directory,
)


class Minato:
    def __init__(self, config: Optional[Config] = None) -> None:
        self._config = config or Config.load()
        self._cache = Cache(
            root=self._config.cache_root,
            expire_days=self._config.expire_days,
        )

    @property
    def cache(self) -> Cache:
        return self._cache

    @contextmanager
    def open(
        self,
        url_or_filename: Union[str, Path],
        mode: str = "r",
        extract: bool = False,
        use_cache: bool = True,
        force_download: bool = False,
        force_extract: bool = False,
    ) -> Iterator[IO[Any]]:
        if (
            not ("a" in mode and "w" in mode and "x" in mode and "+" in mode)
            and use_cache
        ):
            url_or_filename = self.cached_path(
                url_or_filename,
                extract=extract,
                force_download=force_download,
                force_extract=force_extract,
            )

        with open_file(url_or_filename, mode) as fp:
            yield fp

    def cached_path(
        self,
        url_or_filename: Union[str, Path],
        extract: bool = False,
        force_download: bool = False,
        force_extract: bool = False,
        retry: bool = True,
    ) -> Path:
        url_or_filename = str(url_or_filename)

        if "!" in url_or_filename:
            remote_archive_path, file_path = url_or_filename.rsplit("!", 1)
            archive_path = self.cached_path(
                remote_archive_path,
                extract=True,
                force_extract=force_extract,
                force_download=force_download,
            )
            if not archive_path.is_dir():
                raise ValueError(
                    f"{url_or_filename} uses the ! syntax, but this is not an archive file."
                )

            file_path = extract_path(file_path)
            local_path = archive_path / file_path
            if not local_path.exists():
                raise FileNotFoundError(local_path)
            return local_path

        if is_local(url_or_filename):
            url_or_filename = extract_path(url_or_filename)
            if not extract and not is_archive_file(url_or_filename):
                local_path = Path(url_or_filename)
                if not local_path.exists():
                    raise FileNotFoundError(local_path)
                return local_path

        url = str(url_or_filename)
        if url in self._cache:
            cached_file = self._cache.by_url(url)
        else:
            cached_file = self._cache.new(url)

        with self._cache.lock(cached_file):
            if not self._cache.exists(cached_file):
                self._cache.add(cached_file)

            cached_file = self._cache.by_url(url)

            if retry and cached_file.status != CacheStatus.COMPLETED:
                force_download = True

            try:
                downloaded = False
                if (
                    not cached_file.local_path.exists()
                    or self._cache.is_expired(cached_file)
                    or force_download
                ):
                    remove_file_or_directory(cached_file.local_path)

                    cached_file.status = CacheStatus.DOWNLOADING
                    self._cache.update(cached_file)

                    self.download(cached_file.url, cached_file.local_path)
                    downloaded = True

                extracted = False
                if (
                    (extract and cached_file.extraction_path is None)
                    or (downloaded and cached_file.extraction_path is not None)
                    or force_extract
                ) and is_archive_file(cached_file.local_path):
                    cached_file.extraction_path = Path(
                        str(cached_file.local_path) + "-extracted"
                    )
                    remove_file_or_directory(cached_file.extraction_path)

                    cached_file.status = CacheStatus.EXTRACTING
                    self._cache.update(cached_file)

                    extract_archive_file(
                        cached_file.local_path, cached_file.extraction_path
                    )
                    extracted = True

                if downloaded or extracted:
                    cached_file.status = CacheStatus.COMPLETED
                    self._cache.update(cached_file)
            except FileNotFoundError:
                self._cache.delete(cached_file)
                raise
            except (Exception, SystemExit, KeyboardInterrupt):
                cached_file.status = CacheStatus.FAILED
                self._cache.update(cached_file)
                raise

        if (extract or force_extract) and cached_file.extraction_path:
            if not cached_file.extraction_path.exists():
                raise FileNotFoundError(cached_file.extraction_path)
            if cached_file.status != CacheStatus.COMPLETED:
                raise InvalidCacheStatus(
                    f"Cached path status is not completed: status={cached_file.status}"
                )
            return cached_file.extraction_path

        if not cached_file.local_path.exists():
            raise FileNotFoundError(cached_file.local_path)
        if cached_file.status != CacheStatus.COMPLETED:
            raise InvalidCacheStatus(
                f"Cached path status is not completed: status={cached_file.status}"
            )
        return cached_file.local_path

    @staticmethod
    def download(url: str, filename: Path) -> None:
        download(url, filename)

    @staticmethod
    def upload(filename: Path, url: str) -> None:
        with open(filename, "rb") as local_file:
            with open_file(url, "wb") as remote_file:
                content = local_file.read()
                remote_file.write(content)

    def remove(self, url: str) -> None:
        cached_file = self._cache.by_url(url)

        remove_file_or_directory(cached_file.local_path)
        if cached_file.extraction_path is not None:
            remove_file_or_directory(cached_file.extraction_path)
        self._cache.delete(cached_file)



from __future__ import annotations

from contextlib import contextmanager
from pathlib import Path
from typing import IO, Any, Callable, Dict, Iterator, List, Type, Union
from urllib.parse import urlparse

from minato.url import URL


@contextmanager
def open_file(
    url_or_filename: Union[str, Path],
    mode: str = "r",
) -> Iterator[IO[Any]]:

    url = str(url_or_filename)
    filesystem = FileSystem.by_url(url)
    with filesystem.open_file(mode) as fp:
        yield fp


def download(
    url_or_filename: Union[str, Path],
    download_path: Union[str, Path],
) -> None:
    url = str(url_or_filename)
    filesystem = FileSystem.by_url(url)
    filesystem.download(download_path)


class FileSystem:
    registry: Dict[str, Type["FileSystem"]] = {}

    @classmethod
    def register(
        cls, schemes: List[str]
    ) -> Callable[[Type[FileSystem]], Type[FileSystem]]:
        def decorator(subclass: Type[FileSystem]) -> Type[FileSystem]:
            for scheme in schemes:
                FileSystem.registry[scheme] = subclass
            return subclass

        return decorator

    @classmethod
    def by_url(cls, url: str) -> FileSystem:
        parsed_url = urlparse(url)
        scheme = parsed_url.scheme
        if scheme not in FileSystem.registry:
            schemes = ", ".join(FileSystem.registry.keys())
            raise KeyError(f"Invalid scheme: {scheme} (not in {schemes})")
        subclass = FileSystem.registry[scheme]
        return subclass(url)

    def __init__(self, url_or_filename: Union[str, Path]) -> None:
        self._url = URL(str(url_or_filename))

    def exists(self) -> bool:
        raise NotImplementedError

    def download(self, path: Union[str, Path]) -> None:
        raise NotImplementedError

    @contextmanager
    def open_file(self, mode: str = "r") -> Iterator[IO[Any]]:
        raise NotImplementedError



from __future__ import annotations

from typing import Any, Dict, List, Optional


class Table:
    _MIN_MAX_COLUMN_WIDTH = 3

    def __init__(
        self,
        columns: List[str],
        max_column_width: Optional[int] = None,
    ) -> None:
        if (
            max_column_width is not None
            and max_column_width < Table._MIN_MAX_COLUMN_WIDTH
        ):
            raise ValueError(
                f"max_column_widh must be larger than"
                f"{Table._MIN_MAX_COLUMN_WIDTH}: {max_column_width}"
            )

        self._columns = columns
        self._max_column_width = max_column_width
        self._items: List[Dict[str, Any]] = []

    def __getitem__(self, columns: List[str]) -> Table:
        table = Table(columns=columns, max_column_width=self._max_column_width)
        for item in self._items:
            table.add(item)
        return table

    @staticmethod
    def _get_padded_column_value(value: str, width: int) -> str:
        if len(value) > width:
            assert len(value) >= Table._MIN_MAX_COLUMN_WIDTH
            value = value[: width - 2] + ".."

        return f"{value:{width}}"

    @staticmethod
    def _get_column_value_str(value: Any) -> str:
        if isinstance(value, str):
            return repr(value)[1:-1]
        return repr(value)

    def _get_column_widths(self) -> Dict[str, int]:
        column_widths: Dict[str, int] = {}
        for col in self.columns:
            column_values = [x[col] for x in self._items]
            column_value_strings = [
                self._get_column_value_str(x) for x in column_values
            ]

            column_width = max(len(x) for x in column_value_strings + [col])

            if self._max_column_width is not None:
                column_width = min(column_width, self._max_column_width)

            column_widths[col] = column_width
        return column_widths

    @property
    def columns(self) -> List[str]:
        return self._columns

    def add(self, item: Dict[str, Any]) -> None:
        self._items.append({col: item[col] for col in self.columns})

    def sort(self, column: str, desc: bool = False) -> None:
        def _key(item: Dict[str, Any]) -> Any:
            return item[column]

        self._items = sorted(self._items, key=_key, reverse=desc)

    def print(self) -> None:
        column_widths = self._get_column_widths()

        print(
            " ".join(
                self._get_padded_column_value(col, column_widths[col])
                for col in self.columns
            )
        )
        print(" ".join("=" * column_widths[col] for col in self.columns))
        for item in self._items:
            print(
                " ".join(
                    self._get_padded_column_value(
                        self._get_column_value_str(item[col]), column_widths[col]
                    )
                    for col in self.columns
                )
            )



import argparse
from pathlib import Path

from minato.commands.subcommand import Subcommand
from minato.minato import Minato


@Subcommand.register(
    "upload",
    description="upload local file to remote",
    help="upload local file to remote",
)
class UploadCommand(Subcommand):
    def set_arguments(self) -> None:
        # TODO: id or url
        self.parser.add_argument("local", type=Path)
        self.parser.add_argument("remote", type=str)

    def run(self, args: argparse.Namespace) -> None:
        print(f"upload {args.local} to {args.remote}")
        Minato.upload(args.local, args.remote)



import argparse
from pathlib import Path

from minato.commands.subcommand import Subcommand
from minato.minato import Minato
from minato.util import get_parent_path_and_filename


@Subcommand.register(
    "download",
    description="download file to local",
    help="download file to local",
)
class DownloadCommand(Subcommand):
    def set_arguments(self) -> None:
        self.parser.add_argument("url", type=str)
        self.parser.add_argument("path", type=Path)
        self.parser.add_argument("--overwrite", action="store_true")

    def run(self, args: argparse.Namespace) -> None:
        url = args.url
        file_path = args.path
        if file_path.is_dir():
            _, filename = get_parent_path_and_filename(args.url)
            file_path = file_path / filename

        if file_path.exists() and not args.overwrite:
            raise FileExistsError(f"File {str(file_path)} already exists.")

        Minato.download(url, file_path)



import os
import re
import tempfile
from contextlib import contextmanager
from pathlib import Path
from typing import IO, Any, Iterator, Union

from google.cloud.storage import Blob, Client

from minato.filesystems.filesystem import FileSystem


@FileSystem.register(["gs", "gcs"])
class GCSFileSystem(FileSystem):
    def __init__(self, url_or_filename: Union[str, Path]) -> None:
        super().__init__(url_or_filename)
        self._bucket_name = self._url.hostname or ""
        self._key = re.sub(r"^/", "", self._url.path)

        self._project = self._url.get_query("project")
        self._api_endpoint = self._url.get_query("api_endpoint")

        self._client = Client()
        if self._project:
            self._client.project = self._project
        if self._api_endpoint:
            self._client.client_options = {"api_endpoint": self._api_endpoint}

    def _get_blob(self, blob_name: str) -> Blob:
        bucket = self._client.bucket(self._bucket_name)
        blob = bucket.blob(blob_name)
        return blob

    def _get_url_from_blob(self, blob: Blob) -> str:
        return "gs://{self._bucket_name}/{blob.name}"

    def _download_fileobj(self, key: str, fp: IO[Any]) -> None:
        blob = self._get_blob(key)
        blob.download_to_file(fp)

    def _upload_fileobj(self, fp: IO[Any], key: str) -> None:
        blob = self._get_blob(key)
        blob.upload_from_file(fp)

    def exists(self) -> bool:
        client = self._client
        bucket = client.bucket(self._bucket_name)
        blobs = list(bucket.list_blobs(prefix=self._key))
        return len(blobs) > 0

    def download(self, path: Union[str, Path]) -> None:
        if not self.exists():
            raise FileNotFoundError(self._url.raw)

        if isinstance(path, str):
            path = Path(path)

        client = self._client
        bucket = client.bucket(self._bucket_name)
        blobs = list(bucket.list_blobs(prefix=self._key))

        for blob in blobs:
            relpath = os.path.relpath(blob.name, self._key)
            file_path = path / relpath
            os.makedirs(file_path.parent, exist_ok=True)
            if not blob.name.endswith("/"):
                blob.download_to_filename(str(file_path))

    @contextmanager
    def open_file(
        self,
        mode: str = "r",
    ) -> Iterator[IO[Any]]:
        if "x" in mode and self.exists():
            raise FileExistsError(self._url.raw)

        local_file = tempfile.NamedTemporaryFile(delete=False)
        try:
            if "r" in mode or "a" in mode or "+" in mode:
                if not self.exists():
                    raise FileNotFoundError(self._url.raw)
                self._download_fileobj(self._key, local_file)

            local_file.close()
            with open(local_file.name, mode) as fp:
                yield fp

            if "w" in mode or "a" in mode or "+" in mode or "x" in mode:
                with open(local_file.name, "rb") as fp:
                    self._upload_fileobj(fp, self._key)

        finally:
            local_file.close()
            os.remove(local_file.name)



from __future__ import annotations

import argparse
from typing import Callable, List, Optional, Type


class Subcommand:
    subcommands: List["Subcommand"] = []

    @classmethod
    def register(
        cls, name: str, description: str, help: str
    ) -> Callable[[Type[Subcommand]], Type[Subcommand]]:
        def decorator(T: Type[Subcommand]) -> Type[Subcommand]:
            subcommand = T(name, description, help)
            Subcommand.subcommands.append(subcommand)
            return T

        return decorator

    def __init__(self, name: str, description: str, help_: str) -> None:
        self._name = name
        self._description = description
        self._help = help_

        self._parser: Optional[argparse.ArgumentParser] = None

    @property
    def name(self) -> str:
        return self._name

    @property
    def parser(self) -> argparse.ArgumentParser:
        if self._parser is None:
            raise RuntimeError("parser is not set")
        return self._parser

    def setup(self, subparsers: argparse._SubParsersAction) -> None:
        self._parser = subparsers.add_parser(
            self._name,
            description=self._description,
            help=self._help,
        )
        self.set_arguments()
        self.parser.set_defaults(func=self.run)

    def set_arguments(self) -> None:
        raise NotImplementedError

    def run(self, args: argparse.Namespace) -> None:
        raise NotImplementedError



from __future__ import annotations

import dataclasses
from configparser import ConfigParser
from pathlib import Path
from typing import List, Optional, Union

MINATO_ROOT = Path.home() / ".minato"
DEFAULT_CACHE_ROOT = MINATO_ROOT / "cache"
ROOT_CONFIG_PATH = MINATO_ROOT / "config.ini"
LOCAL_CONFIG_PATH = Path.cwd() / "minato.ini"


@dataclasses.dataclass
class Config:
    cache_root: Path = DEFAULT_CACHE_ROOT
    expire_days: int = -1

    @classmethod
    def load(
        cls,
        cache_root: Optional[Union[str, Path]] = None,
        expire_days: Optional[int] = None,
        files: Optional[List[Union[str, Path]]] = None,
    ) -> Config:
        if files is None:
            files = [ROOT_CONFIG_PATH, LOCAL_CONFIG_PATH]

        config = cls()
        config.read_files(files)
        if cache_root is not None:
            config.cache_root = Path(cache_root)
        if expire_days is not None:
            config.expire_days = expire_days

        return config

    def read_files(self, files: List[Union[str, Path]]) -> None:
        parser = ConfigParser()
        parser.read([str(path) for path in files])
        self._update_from_configparser(parser)

    def _update_from_configparser(self, parser: ConfigParser) -> None:
        if parser.has_section("cache"):
            section = parser["cache"]
            if "root" in section:
                self.cache_root = Path(parser["cache"]["root"])
            if "expire_days" in section:
                self.expire_days = int(parser["cache"]["expire_days"])

    @property
    def cache_db_path(self) -> Path:
        return self.cache_root / "cache.db"

    @property
    def cache_artifact_dir(self) -> Path:
        return self.cache_root / "artifacts"



import argparse
from typing import Optional

from minato import __version__
from minato.commands.cache import CacheCommand  # noqa: F401
from minato.commands.download import DownloadCommand  # noqa: F401
from minato.commands.list import ListCommand  # noqa: F401
from minato.commands.remove import RemoveCommand  # noqa: F401
from minato.commands.subcommand import Subcommand
from minato.commands.update import UpdateCommand  # noqa: F401
from minato.commands.upload import UploadCommand  # noqa: F401


def create_parser(prog: Optional[str] = None) -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(usage="%(prog)s", prog=prog)
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s " + __version__,
    )

    subparsers = parser.add_subparsers()

    for subcommand in Subcommand.subcommands:
        subcommand.setup(subparsers)

    return parser


def main(prog: Optional[str] = None) -> None:
    parser = create_parser(prog)
    args = parser.parse_args()

    func = getattr(args, "func", None)
    if func is None:
        parser.parse_args(["--help"])

    func(args)



import argparse
from pathlib import Path

from minato.commands.subcommand import Subcommand
from minato.config import Config
from minato.minato import Minato


@Subcommand.register(
    "update",
    description="update cached files",
    help="update cached files",
)
class UpdateCommand(Subcommand):
    def set_arguments(self) -> None:
        self.parser.add_argument("uid_or_url", nargs="*", default=[])
        self.parser.add_argument("--force", action="store_true")
        self.parser.add_argument("--expired", action="store_true")
        self.parser.add_argument("--failed", action="store_true")
        self.parser.add_argument("--expire-days", type=int, default=None)
        self.parser.add_argument("--root", type=Path, default=None)

    def run(self, args: argparse.Namespace) -> None:
        config = Config.load(
            cache_root=args.root,
            expire_days=args.expire_days,
        )
        minato = Minato(config)
        cache = minato.cache

        cached_files = cache.filter(
            queries=args.uid_or_url,
            expired=args.expired or args.expire_days is not None,
            failed=args.failed,
        )

        num_caches = len(cached_files)
        if num_caches == 0:
            print("No caches to update.")
            return

        print(f"{num_caches} files will be updated:")
        for cached_file in cached_files:
            print(f"  [{cached_file.uid[:8]}] {cached_file.url}")

        if not args.force:
            yes_or_no = input("Are you sure to update these caches? y/[n]:")
            if yes_or_no not in ("y", "Y"):
                print("canceled")
                return

        for cached_file in cached_files:
            minato.cached_path(cached_file.url)

        print("Cache files were successfully updated.")



import argparse
from pathlib import Path

from minato.commands.subcommand import Subcommand
from minato.config import Config
from minato.minato import Minato


@Subcommand.register(
    "cache",
    description="cache remote file and return cached local file path",
    help="cache remote file and return cached local file path",
)
class CacheCommand(Subcommand):
    def set_arguments(self) -> None:
        self.parser.add_argument("url", type=str)
        self.parser.add_argument("--extract", action="store_true")
        self.parser.add_argument("--force-extract", action="store_true")
        self.parser.add_argument("--force-download", action="store_true")
        self.parser.add_argument("--not-retry", action="store_false")
        self.parser.add_argument("--expire-days", type=int, default=None)
        self.parser.add_argument("--root", type=Path, default=None)

    def run(self, args: argparse.Namespace) -> None:
        config = Config.load(
            cache_root=args.root,
            expire_days=args.expire_days,
        )
        minato = Minato(config)

        cached_path = minato.cached_path(
            args.url,
            extract=args.extract,
            force_extract=args.force_extract,
            force_download=args.force_download,
            retry=args.not_retry,
        )
        print(cached_path)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['minato', 'minato.commands', 'minato.filesystems']

package_data = \
{'': ['*']}

install_requires = \
['boto3>=1.17.97,<2.0.0',
 'google-cloud-storage>=1.39.0,<2.0.0',
 'requests>=2.25.1,<3.0.0',
 'tqdm>=4.60.0,<5.0.0']

entry_points = \
{'console_scripts': ['minato = minato.__main__:run']}

setup_kwargs = {
    'name': 'minato',
    'version': '0.3.0',
    'description': 'Cache and file system for online resources',
    'long_description': 'Minato\n======\n\n[![Actions Status](https://github.com/altescy/minato/workflows/CI/badge.svg)](https://github.com/altescy/minato/actions/workflows/main.yml)\n[![Python version](https://img.shields.io/pypi/pyversions/minato)](https://github.com/altescy/minato)\n[![License](https://img.shields.io/github/license/altescy/minato)](https://github.com/altescy/minato/blob/master/LICENSE)\n[![pypi version](https://img.shields.io/pypi/v/minato)](https://pypi.org/project/minato/)\n\nCache & file system for online resources in Python\n\n\n## Installation\n\n```\n$ pip install minato\n```\n\n## Usage\n\n### Python\n\n```python\nimport minato\n\n# Read / write files on online storage\nwith minato.open("s3://your_bucket/path/to/file", "w") as f:\n    f.write("Create a new file on AWS S3!")\n\n# Cache & manage online resources in local storage\nlocal_filename = minato.cached_path("http://example.com/path/to/archive.zip!inner/path/to/file")\n```\n\n### CLI\n\n```\n$ minato --help\nusage: minato\n\npositional arguments:\n  {cache,download,list,remove,update,upload}\n    cache               cache remote file and return cached local file path\n    download            download file to local\n    list                show list of cached files\n    remove              remove cached files\n    update              update cached files\n    upload              upload local file to remote\n\noptional arguments:\n  -h, --help            show this help message and exit\n  --version             show program\'s version number and exit\n```\n',
    'author': 'Yasuhiro Yamaguchi',
    'author_email': 'altescy@fastmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/altescy/minato',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import dataclasses
import datetime
import json
import os
import uuid
from contextlib import contextmanager
from enum import Enum
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union

from minato.exceptions import CacheAlreadyExists, CacheNotFoundError, ConfigurationError
from minato.filelock import FileLock
from minato.util import remove_file_or_directory


class CacheStatus(Enum):
    PENDING = "PENDING"
    DOWNLOADING = "DOWNLOADING"
    EXTRACTING = "EXTRACTING"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    DELETED = "DELETED"


@dataclasses.dataclass
class CachedFile:
    uid: str
    url: str
    local_path: Path
    created_at: datetime.datetime
    updated_at: datetime.datetime
    extraction_path: Optional[Path]
    status: CacheStatus

    def __init__(
        self,
        uid: str,
        url: str,
        local_path: Union[str, Path],
        created_at: Union[str, datetime.datetime],
        updated_at: Union[str, datetime.datetime],
        extraction_path: Optional[Union[str, Path]] = None,
        status: Union[str, CacheStatus] = CacheStatus.PENDING,
    ) -> None:
        if isinstance(local_path, str):
            local_path = Path(local_path)
        if isinstance(created_at, str):
            created_at = datetime.datetime.fromisoformat(created_at)
        if isinstance(updated_at, str):
            updated_at = datetime.datetime.fromisoformat(updated_at)
        if isinstance(extraction_path, str):
            extraction_path = Path(extraction_path)
        if extraction_path is not None:
            extraction_path = extraction_path.absolute()
        if isinstance(status, str):
            status = CacheStatus(status)

        self.uid = uid
        self.url = url
        self.local_path = local_path.absolute()
        self.created_at = created_at
        self.updated_at = updated_at
        self.extraction_path = extraction_path
        self.status = status

    def to_tuple(self) -> Tuple[str, str, str, str, str, Optional[str], str]:
        return (
            self.uid,
            self.url,
            str(self.local_path),
            self.created_at.isoformat(),
            self.updated_at.isoformat(),
            str(self.extraction_path) if self.extraction_path else None,
            self.status.value,
        )

    def to_dict(self) -> Dict[str, Any]:
        return {
            "uid": str(self.uid),
            "url": str(self.url),
            "local_path": str(self.local_path),
            "created_at": self.created_at.isoformat(),
            "updated_at": self.updated_at.isoformat(),
            "extraction_path": str(self.extraction_path)
            if self.extraction_path
            else None,
            "status": self.status.value,
        }


class Cache:
    def __init__(
        self,
        root: Path,
        expire_days: int = -1,
    ) -> None:
        if not root.exists():
            os.makedirs(root, exist_ok=True)

        if not root.is_dir():
            raise ConfigurationError(
                f"Given cache_directory path is not a directory: {root}"
            )

        self._root = root
        self._expire_days = expire_days

    def __contains__(self, url: str) -> bool:
        try:
            self.by_url(url)
            return True
        except CacheNotFoundError:
            return False

    @contextmanager
    def lock(self, item: CachedFile) -> Iterator[None]:
        lock = FileLock(self.get_lockfile_path(item.uid))
        try:
            lock.acquire()
            yield
        finally:
            lock.release()

    @staticmethod
    def _generate_uid() -> str:
        return uuid.uuid4().hex

    def get_metadata_path(self, uid: str) -> Path:
        return self._root / (uid + ".json")

    def get_lockfile_path(self, uid: str) -> Path:
        return self._root / (uid + ".lock")

    def load_cached_file(self, metadata_path: Path) -> CachedFile:
        if not metadata_path.exists():
            raise CacheNotFoundError(f"Cache not found: {metadata_path}")
        with open(metadata_path, "r") as fp:
            params = json.load(fp)
        return CachedFile(**params)

    def new(self, url: str) -> CachedFile:
        uid = self._generate_uid()
        cached_file = CachedFile(
            uid=uid,
            url=url,
            local_path=self._root / uid,
            created_at=datetime.datetime.now(),
            updated_at=datetime.datetime.now(),
        )
        return cached_file

    def exists(self, item: CachedFile) -> bool:
        metadata_path = self.get_metadata_path(item.uid)
        return metadata_path.exists()

    def save(self, item: CachedFile) -> None:
        metadata_path = self.get_metadata_path(item.uid)
        with open(metadata_path, "w") as fp:
            json.dump(item.to_dict(), fp)

    def add(self, item: CachedFile) -> CachedFile:
        if self.exists(item):
            raise CacheAlreadyExists(item.url)
        self.save(item)
        return item

    def update(self, item: CachedFile) -> None:
        metadata_path = self.get_metadata_path(item.uid)
        if not metadata_path.exists():
            raise CacheNotFoundError(f"Cache not found with uid={item.uid}")
        item.updated_at = datetime.datetime.now()
        self.save(item)

    def by_uid(self, uid: str) -> CachedFile:
        metadata_path = self.get_metadata_path(uid)
        if not metadata_path.exists():
            raise CacheNotFoundError(f"Cache not found with uid={uid}")
        with open(metadata_path, "r") as fp:
            params = json.load(fp)
        return CachedFile(**params)

    def by_url(self, url: str) -> CachedFile:
        for cached_file in self.all():
            if cached_file.url == url:
                return cached_file
        raise CacheNotFoundError(f"Cache not found with url={url}")

    def delete(self, item: CachedFile) -> None:
        metadata_path = self.get_metadata_path(item.uid)
        lockfile_path = self.get_lockfile_path(item.uid)

        remove_file_or_directory(item.local_path)
        if item.extraction_path is not None:
            remove_file_or_directory(item.extraction_path)

        remove_file_or_directory(metadata_path)
        remove_file_or_directory(lockfile_path)

    def is_expired(self, item: CachedFile) -> bool:
        if self._expire_days < 0:
            return False
        now = datetime.datetime.now()
        delta = now - item.updated_at
        return delta.days >= self._expire_days

    def all(self) -> List[CachedFile]:
        cached_files: List[CachedFile] = []
        for metafile_path in self._root.glob("*.json"):
            with open(metafile_path, "r") as fp:
                params = json.load(fp)
                cached_file = CachedFile(**params)
            cached_files.append(cached_file)
        cached_files = sorted(cached_files, key=lambda x: x.created_at)
        return cached_files

    def filter(
        self,
        queries: List[str],
        expired: Optional[bool] = None,
        failed: Optional[bool] = None,
        completed: Optional[bool] = None,
    ) -> List[CachedFile]:
        cached_files = self.all()
        for query in queries:
            cached_files = [
                x for x in cached_files if query in x.url or x.uid.startswith(query)
            ]
        if expired is not None:
            cached_files = [x for x in cached_files if self.is_expired(x) == expired]
        if failed is not None:
            cached_files = [
                x for x in cached_files if (x.status == CacheStatus.FAILED) == failed
            ]
        if completed is not None:
            cached_files = [
                x
                for x in cached_files
                if (x.status == CacheStatus.COMPLETED) == completed
            ]

        unique_caches = {x.uid: x for x in cached_files}
        cached_files = sorted(unique_caches.values(), key=lambda x: x.created_at)
        return cached_files



import os
import tempfile
from contextlib import contextmanager
from pathlib import Path
from typing import IO, Any, Iterator, Union

import requests

from minato.filesystems.filesystem import FileSystem
from minato.util import http_get


@FileSystem.register(["http", "https"])
class HttpFileSystem(FileSystem):
    def exists(self) -> bool:
        response = requests.head(self._url.raw, allow_redirects=True)
        status_code = response.status_code
        return status_code == 200

    def download(self, path: Union[str, Path]) -> None:
        if not self.exists():
            raise FileNotFoundError(self._url.raw)

        with open(path, "w+b") as fp:
            http_get(self._url.raw, fp)

    @contextmanager
    def open_file(
        self,
        mode: str = "r",
    ) -> Iterator[IO[Any]]:
        if not self.exists():
            raise FileNotFoundError(self._url.raw)

        if "a" in mode or "w" in mode or "+" in mode or "x" in mode:
            raise ValueError("HttpFileSystem is not writable.")

        temp_file = tempfile.NamedTemporaryFile(delete=False)
        try:
            http_get(self._url.raw, temp_file)
            temp_file.close()
            with open(temp_file.name, mode) as fp:
                yield fp
        finally:
            os.remove(temp_file.name)
